from typing import Dict, List, Literal, Optional
from django.conf import settings
from django.db.models import Model
from rest_framework.serializers import Serializer
from rudra.settings import RudraBaseSettings
from django.http import HttpRequest
from rudra.paginator import PaginatorService


METHODS = Literal['get', 'post', 'put', 'patch', 'delete', 'head', 'options', 'trace']
rudrasettings: RudraBaseSettings = settings.RUDRASETTINGS

def get_class(kls):
    parts = kls.split('.')
    module = ".".join(parts[:-1])
    m = __import__( module )
    for comp in parts[1:]:
        m = getattr(m, comp)            
    return m

def get_success_response(data: dict = None, paginator: PaginatorService = None):
    success_path = rudrasettings.success_path
    success = get_class(success_path)
    if paginator:
        return success(data, paginator.has_next, paginator.pages, paginator.get_query_set_count, paginator.page_size)
    return success(data)

def get_error_response(message: str):
    error_path = rudrasettings.error_path
    error = get_class(error_path)
    return error(Exception(message))

def get_models() -> List[Model]:
    return [get_class(meta_settings.model_path) for meta_settings in rudrasettings.meta_settings]

def get_model(model_name: str) -> Optional[Model]:
    for meta_settings in rudrasettings.meta_settings:
        if model_name.lower() == meta_settings.model_path.split('.')[-1].lower():
            return get_class(meta_settings.model_path)
    return None

def get_serializer(model: Model, method: METHODS = None) -> Optional[Serializer]:
    for meta_settings in rudrasettings.meta_settings:
        if meta_settings.model_path.split('.')[-1].lower() == model.__name__.lower():
            if meta_settings.serializer_settings.serializer_path:
                return get_class(meta_settings.serializer_settings.serializer_path)
            else:
                if method:
                    return get_class(meta_settings.serializer_settings.map_serializer.get(method))

def get_user_serializer() -> Optional[Serializer]:
    if rudrasettings.user_serializer_path:
        return get_class(rudrasettings.user_serializer_path)
    return None

def check_method_allowed(model: Model, method: METHODS) -> bool:
    for meta_settings in rudrasettings.meta_settings:
        if meta_settings.model_path.split('.')[-1].lower() == model.__name__.lower():
            return method in meta_settings.methods_allowed
    return False

def get_serializer_context(request: HttpRequest) -> Dict:
    for meta_settings in rudrasettings.meta_settings:
        if meta_settings.serializer_settings.serializer_context:
            return meta_settings.serializer_settings.serializer_context(request)
    return {}


from functools import reduce
import operator
from typing import Optional
from django.http import HttpRequest
from rest_framework.views import APIView

from rudra.paginator import PaginatorService
from rudra.service import get_model, get_models, get_serializer, get_success_response, get_error_response, check_method_allowed, get_serializer_context, get_user_serializer
from rest_framework.serializers import Serializer
from django.db.models import Q

class AllModels(APIView):
    def get(self, request: HttpRequest):

        models = get_models()

        models_with_fields = {}

        for model in models:
            models_with_fields[model.__name__] = [{
                'name': field.name,
                'type': field.get_internal_type(),
                'related_model': field.related_model.__name__ if field.related_model else None,
                'description': field.description if hasattr(field, 'description') else None
            } for field in model._meta.get_fields()]

        return get_success_response(models_with_fields)

class User(APIView):
    def get(self, request: HttpRequest):
        user_serializer = get_user_serializer()
        if user_serializer:
            if not request.user.is_authenticated:
                return get_error_response('User not authenticated')
            return get_success_response(user_serializer(request.user).data)
        return get_error_response('User serializer not found')

class QueryModel(APIView):
    def dispatch(self, request: HttpRequest, *args, **kwargs):
        method = request.method.lower()
        model_name = kwargs.get('model_name')
        model = get_model(model_name)

        if not model:
            return get_error_response('Model not found')

        if not check_method_allowed(model, method):
            return get_error_response('Method not allowed')
        
        self.model = model        
        self.serializer = get_serializer(model, method=method)
        self.serializer_context = get_serializer_context(request)

        return super().dispatch(request, *args, **kwargs)
    
    def is_valid_data(self, serializer: Serializer) -> Optional[str]:
        try:
            serializer.is_valid(raise_exception=True)
        except Exception as e:
            return str(e)

    def get(self, request: HttpRequest, model_name: str):
        filters = request.GET.dict()
        all = bool(filters.pop('all', False))
        page = filters.pop('page', None)
        page_size = filters.pop('page_size', None)
        order_by = filters.pop('order_by', None)
        instance = self.model.objects.filter(**filters)

        if all:
            instance = instance.all()
            return get_success_response(self.serializer(instance, many=True, context=self.serializer_context).data)
        elif page:
            if not order_by:
                order_by = '-id'
            instance = instance.order_by(order_by)
            paginator = PaginatorService(instance, page=page or 1, page_size=page_size or 10)
            return get_success_response(self.serializer(paginator.get_query_set, many=True, context=self.serializer_context).data, paginator)
        else:
            instance = instance.first()
            if not instance:
                return get_error_response('Not found')
            return get_success_response(self.serializer(instance, context=self.serializer_context).data)

    def post(self, request: HttpRequest, model_name: str):
        data = request.data
        serializer = self.serializer(data=data, context=self.serializer_context)
        error = self.is_valid_data(serializer)
        if error:
            return get_error_response(error)
        instance = serializer.save()
        return get_success_response(self.serializer(instance, context=self.serializer_context).data)
    
    def put(self, request: HttpRequest, model_name: str):
        data = request.data
        serializer = self.serializer(data=data, context=self.serializer_context)
        error = self.is_valid_data(serializer)
        if error:
            return get_error_response(error)
        instance = serializer.save()
        return get_success_response(self.serializer(instance, context=self.serializer_context).data)
    
    def patch(self, request: HttpRequest, model_name: str):
        data = request.data
        serializer = self.serializer(data=data, context=self.serializer_context)
        error = self.is_valid_data(serializer)
        if error:
            return get_error_response(error)
        instance = serializer.save()
        return get_success_response(self.serializer(instance, context=self.serializer_context).data)
    
    def delete(self, request: HttpRequest, model_name: str):
        data = request.data
        instance = self.model.objects.filter(**data).first()
        if not instance:
            return get_error_response('Instance not found')
        instance.delete()
        return get_success_response()

class DeepQueryModel(APIView):

    def dispatch(self, request: HttpRequest, *args, **kwargs):
        method = request.method.lower()
        model_name = kwargs.get('model_name')
        model = get_model(model_name)

        if not model:
            return get_error_response('Model not found')

        if not check_method_allowed(model, method):
            return get_error_response('Method not allowed')
        
        self.model = model        
        self.serializer = get_serializer(model, method=method)
        self.serializer_context = get_serializer_context(request)

        return super().dispatch(request, *args, **kwargs)

    def post(self, request: HttpRequest, model_name: str):
        data = request.data
        query_params = request.GET.dict()
        model = get_model(model_name)
        if not model:
            return get_error_response('Model not found')

        object_manager = model.objects

        # Select related query
        select_related = data.get('select_related', [])
        if select_related:
            object_manager = object_manager.select_related(*select_related)

        # Prefetch related query 
        prefetch_related = data.get('prefetch_related', [])
        if prefetch_related:
            object_manager = object_manager.prefetch_related(*prefetch_related)
        
        instance = object_manager

        or_filters = data.get('or_filters', {})
        if or_filters:
            instance = instance.filter(
                reduce(operator.or_, (Q(**{key: value}) for key, value in or_filters.items()))
            )
        
        filters = data.get('filters', {})
        instance = instance.filter(**filters)

        order_by_list = data.get('order_by_list', [])

        if order_by_list:
            instance = instance.order_by(*order_by_list)
        
        all = bool(query_params.get('all', False))
        page = query_params.get('page', None)
        page_size = query_params.get('page_size', None)
        print(instance.query)

        if all:
            instance = instance.all()
            return get_success_response(self.serializer(instance, many=True, context=self.serializer_context).data)
        elif page:
            if not order_by_list:
                order_by_list = ['-id']
            instance = instance.order_by(*order_by_list)
            paginator = PaginatorService(instance, page=page or 1, page_size=page_size or 10)
            return get_success_response(self.serializer(paginator.get_query_set, many=True, context=self.serializer_context).data, paginator)
        else:
            instance = instance.first()
            if not instance:
                return get_error_response('Instance not found')
            return get_success_response(self.serializer(instance, context=self.serializer_context).data)


from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.db.models import QuerySet
from rudra.settings import RudraBaseSettings
from django.conf import settings

rudrasettings: RudraBaseSettings = settings.RUDRASETTINGS

class PaginatorService:

    def __init__(self, queryset: QuerySet, page: int=1, page_size: int=10, orphans: int=0, use_fixed_page_number: bool=True):
        self.queryset = queryset
        self.page_size = int(f"{page_size}")
        self.page = int(f"{page}")
        self.orphans = orphans
        self.page_obj = None
        self.use_fixed_page_number = use_fixed_page_number
        self.fixed_page_number = rudrasettings.fixed_page_number
    
    @property
    def get_page(self):
        if self.page_obj:
            return self.page_obj

        paginator = Paginator(self.queryset, self.page_size, orphans=self.orphans)
        try:
            page_obj = paginator.page(self.page)
        except PageNotAnInteger:
            page_obj = paginator.page(1)
        except EmptyPage:
            page_obj = paginator.page(paginator.num_pages)

        self.page_obj = page_obj
        return self.page_obj
    
    @property
    def get_query_set(self):
        return self.get_page.object_list
    
    @property
    def get_query_set_count(self):
        if self.use_fixed_page_number and self.fixed_page_number:
            return self.fixed_page_number * self.page_size
        return self.get_page.object_list.count()
    
    @property
    def has_next(self):
        if self.use_fixed_page_number and self.fixed_page_number:
            return True
        return self.get_page.has_next()
    
    @property
    def pages(self):
        if self.use_fixed_page_number and self.fixed_page_number:
            return self.fixed_page_number
        return self.get_page.paginator.num_pages


from typing import Callable, Dict, List, Optional
from django.http import HttpRequest

 
class SerializerSettings:
    serializer_path: str
    map_serializer: Dict[str, str]
    serializer_context: Callable[[HttpRequest], dict]

    def __init__(self, serializer_path: str = None, map_serializer: Dict[str, str] = None, serializer_context: Callable[[HttpRequest], dict] = None,):
        self.serializer_path = serializer_path
        self.map_serializer = map_serializer
        self.serializer_context = serializer_context

class RudraMetaSettings:
    methods_allowed: List[str]
    serializer_path: str
    model_path: str
    serializer_settings: SerializerSettings

    def __init__(self, methods_allowed: List[str], model_path: str, serializer_settings: SerializerSettings = None):
        self.methods_allowed = methods_allowed
        self.model_path = model_path 
        self.serializer_settings = serializer_settings

class RudraBaseSettings:
    success_path: str
    error_path: str
    user_serializer_path: str
    fixed_page_number: Optional[int] # for pagination page number
    meta_settings: List[RudraMetaSettings]


from django.urls import path
from rudra import views

urlpatterns = [
    path('get-user/', views.User.as_view()),
    path('models/', views.AllModels.as_view()),
    path('<str:model_name>/', views.QueryModel.as_view()),
    path('query/<str:model_name>/', views.DeepQueryModel.as_view())
]
